'''
Notices:
Copyright 2018 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRessED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNess FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLess THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.
'''

import numpy as np

from .particles import Particles


class Updater:
    '''
    Updates particle weights (and resamples if necessary) based on delta phi
    and particle state.
    '''
    def __init__(self, ess_threshold):
        '''
        :param ess_threshold: threshold on effective sample size (ess); if
            ess < ess_threshold, resampling with replacement is conducted.
        :type ess_threshold: float
        '''
        self.ess_threshold = ess_threshold
        self._ess = np.nan
        self._resampled = False
        self._unnorm_log_weights = []

    @property
    def ess_threshold(self):
        return self._ess_threshold

    @ess_threshold.setter
    def ess_threshold(self, ess_threshold):
        if ess_threshold < 0. or ess_threshold > 1.:
            raise ValueError('ESS threshold must be between 0 and 1.')
        self._ess_threshold = ess_threshold

    @property
    def ess(self):
        return self._ess

    @property
    def resampled(self):
        return self._resampled

    def update(self, particles, delta_phi):
        new_log_weights = self._compute_new_weights(particles, delta_phi)
        new_particles = Particles(particles.param_dict, particles.log_likes,
                                  new_log_weights)

        new_particles = self.resample_if_needed(new_particles)
        return new_particles

    def resample_if_needed(self, particles):
        eff_sample_size = particles.compute_ess()
        self._ess = eff_sample_size
        self._resampled = False

        if eff_sample_size < self.ess_threshold * particles.num_particles:
            self._resampled = True
            resampled_particles = self._resample(particles)
            resampled_particles._total_unlw = particles.total_unnorm_log_weight

            return resampled_particles

        return particles

    def _compute_new_weights(self, particles, delta_phi):
        un_log_weights = particles.log_weights + particles.log_likes * delta_phi
        self._unnorm_log_weights.append(un_log_weights)
        return un_log_weights

    def _resample(self, particles):
        u = np.random.uniform(0, 1, particles.num_particles)
        resample_indices = np.digitize(u, np.cumsum(particles.weights))

        new_params = particles.params[resample_indices]
        param_dict = dict(zip(particles.param_names, new_params.T))

        new_log_likes = particles.log_likes[resample_indices]

        uniform_weights = [1/particles.num_particles] * particles.num_particles
        new_weights = np.log(uniform_weights)

        return Particles(param_dict, new_log_likes, new_weights)



'''
Notices:
Copyright 2018 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRessED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNess FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLess THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.
'''

import numpy as np
import copy
import functools
from smcpy.utils.checks import Checks


def package_for_user(func):

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        Packages array output in a dictionary with keys = parameter names and
        values = array columns UNLESS kwarg package=False is passed to the
        decorated function.
        """
        if 'package' in kwargs.keys() and kwargs['package'] == False:
            return func(self)

        else:
            outputs = func(self, *args, **kwargs)
            names = self.param_names
            return {name: output for name, output in zip(names, outputs)}

    return wrapper



class Particles(Checks):
    '''
    A container for particles during sequential monte carlo (SMC) sampling.
    '''

    def __init__(self, params, log_likes, log_weights):
        '''
        :param params: model parameters; keys are parameter names and values
            are particle values
        :type params: dictionary of 1D arrays or lists
        :param log_likes: natural log of particle likelihoods
        :type log_likes: 1D array or list
        :param log_weights: natural log of particle weights; weights will be
            automatically normalized when set to avoid potential misuse
        :type log_weights: 1D array or list
        '''
        self._param_names = None
        self._num_particles = None

        self.attrs = dict()

        self._set_params(params)
        self._set_log_likes(log_likes)
        self._set_and_norm_log_weights(log_weights)

        self._total_unlw = self._logsum(log_weights)

    @property
    def params(self):
        return self._params

    def _set_params(self, params):
        if not self._is_dict(params):
            raise TypeError('"params" must be dict of array-like objects.')
        self._param_names = tuple(params.keys())
        self._params = np.vstack(list(params.values())).T
        self._num_particles = self._params.shape[0]

    @property
    def param_dict(self):
        return dict(zip(self._param_names, self._params.T))

    @property
    def log_likes(self):
        return self._log_likes

    def _set_log_likes(self, log_likes):
        log_likes = np.array(log_likes).reshape(-1, 1)
        if log_likes.shape[0] != self._num_particles:
            raise ValueError('log_likes.shape[0] != number particles: '
                             f'{log_likes.shape[0]} != {self._num_particles}')
        self._log_likes = log_likes

    @property
    def total_unnorm_log_weight(self):
        return self._total_unlw

    @property
    def log_weights(self):
        return self._log_weights

    @property
    def weights(self):
        return self._weights

    def _set_and_norm_log_weights(self, log_weights):
        log_weights = np.array(log_weights).reshape(-1, 1)
        if log_weights.shape[0] != self._num_particles:
            raise ValueError('log_weights.shape[0] != number particles: '
                             f'{log_weights.shape[0]} != {self._num_particles}')
        self._log_weights = self._normalize_log_weights(log_weights)
        self._weights = np.exp(self._log_weights)

    @property
    def param_names(self):
        return self._param_names

    @property
    def num_particles(self):
        return self._num_particles

    def _normalize_log_weights(self, log_weights):
        '''
        Normalizes log weights, and then transforms back into log space
        '''
        shifted_weights = np.exp(log_weights - np.max(log_weights))
        normalized_weights = shifted_weights / np.sum(shifted_weights)
        log_zero_weights = np.ones(normalized_weights.shape) * -np.inf
        return np.log(normalized_weights, out=log_zero_weights,
                      where=normalized_weights > 0)

    def copy(self):
        '''
        Returns a copy of the entire step class.
        '''
        return copy.deepcopy(self)

    def compute_ess(self):
        '''
        Computes the effective sample size (ess) of the step based on log weight
        '''
        return 1 / np.sum(self.weights ** 2)

    @package_for_user
    def compute_mean(self):
        '''
        Returns the estimated mean of each parameter.
        '''
        return np.sum(self.params * self.weights, axis=0)

    @package_for_user
    def compute_variance(self):
        '''
        Returns the estimated variance of each parameter. Uses weighted
        sample formula https://en.wikipedia.org/wiki/Sample_mean_and_covariance 
        '''
        means = self.compute_mean(package=False)
        norm = 1 - np.sum(self.weights ** 2)
        return np.sum(self.weights * (self.params - means) ** 2, axis=0) / norm

    @package_for_user
    def compute_std_dev(self):
        '''
        Returns the estimated standard deviation of each parameter. 
        '''
        var = self.compute_variance(package=False)
        return np.sqrt(var)

    def compute_covariance(self):
        '''
        Estimates the covariance matrix.
        '''
        cov = np.cov(self.params.T, ddof=0, aweights=self.weights.flatten())

        if cov.shape == ():
            cov = cov.reshape(1, 1)

        return cov

    @staticmethod
    def _logsum(Z):
        Z = -np.sort(-np.array(Z), axis=0).flatten() # descending
        Z0 = Z[0]
        Z_shifted = Z[1:] - Z0
        return Z0 + np.log(1 + np.sum(np.exp(Z_shifted), axis=0))



from .samplers import FixedSampler, AdaptiveSampler
from .priors import ImproperUniform, InvWishart, ImproperCov
from .log_likelihoods import Normal, MultiSourceNormal, MVNormal
from .log_likelihoods import MVNRandomEffects
from .mcmc.vector_mcmc import VectorMCMC
from .mcmc.parallel_mcmc import ParallelMCMC
from .mcmc.vector_mcmc_kernel import VectorMCMCKernel

__version__ = "0.1.4"


import numpy as np

from .vector_mcmc import VectorMCMC
from ..log_likelihoods import Normal


class ParallelMCMC(VectorMCMC):
    '''
    Enables use of MPI to split model evaluations over a distributed memory
    system.

    ParallelMCMC is set up such that all positive ranks have effectively
    garbage values. These garbage values serve the purpose of allowing the
    MCMCBase class, which was not written with MPI in mind, to function
    normally and propagate objects of the right shape and type. At the end of
    an analysis that uses the ParallelMCMC class, outputs on positive ranks
    should be discarded and ONLY the output from rank 0 should be used.
    '''

    def __init__(self, model, data, priors, mpi_comm, log_like_args=None,
                 log_like_func=Normal):
        self._comm = mpi_comm
        self._size = mpi_comm.Get_size()
        self._rank = mpi_comm.Get_rank()

        super().__init__(model, data, priors, log_like_args, log_like_func)

        self._log_like_func.set_model_wrapper(self.model_wrapper)

        try:
            self._output_dim = self._data.shape[1]
        except:
            self._output_dim = self._data.size

    def model_wrapper(self, model, inputs):
        partitioned_inputs = np.array_split(inputs, self._size)
        scattered_inputs = []
        scattered_inputs = self._comm.scatter(partitioned_inputs, root=0)

        scattered_outputs = np.array([]).reshape(0, self._output_dim)
        if scattered_inputs.shape[0] > 0:
            scattered_outputs = model(scattered_inputs)

        gathered_outputs = self._comm.allgather(scattered_outputs)
        outputs = np.concatenate(gathered_outputs)
        return outputs



import numpy as np
import warnings

from abc import ABC, abstractmethod
from tqdm import tqdm

from ..utils.mpi_utils import rank_zero_output_only


class MCMCBase(ABC):
    
    def __init__(self, model, data, priors, log_like_args, log_like_func):
        '''
        :param model: maps inputs to outputs
        :type model: callable
        :param data: data corresponding to model outputs
        :type data: 1D array
        :param priors: random variable objects with a pdf and rvs method (e.g.
            scipy stats random variable objects)
        :type priors: list of objects
        :param log_like_args: any fixed parameters that define the likelihood
            function (e.g., standard deviation for a Gaussian likelihood).
        :type log_like_args: 1D array or None
        :param log_like_func: log likelihood function that takes inputs, model,
            data, and hyperparameters and returns log likelihoods
        :type log_like_func: callable
        '''
        self._eval_model = model
        self._data = data
        self._priors = priors
        self._log_like_func = log_like_func(self.evaluate_model, data,
                                            log_like_args)

    def smc_metropolis(self, inputs, num_samples, cov, phi):
        num_particles = inputs.shape[0]
        log_priors, log_like = self._initialize_probabilities(inputs)

        for i in range(num_samples):

            inputs, log_like, log_priors, rejected = \
                self._perform_mcmc_step(inputs, cov, log_like, log_priors, phi)

            num_accepted = num_particles - np.sum(rejected)

            if num_accepted < inputs.shape[0] * 0.3:
                cov = cov * 1/5
            if num_accepted > inputs.shape[0] * 0.7:
                cov = cov * 2

        return inputs, log_like

    def metropolis(self, inputs, num_samples, cov, adapt_interval=None,
                   adapt_delay=0, progress_bar=False, **kwargs):

        chain = np.zeros([inputs.shape[0], inputs.shape[1], num_samples + 1])
        chain[:, :, 0] = inputs

        log_priors, log_like = self._initialize_probabilities(inputs)

        for i in tqdm(range(1, num_samples + 1), disable=not progress_bar):
    
            inputs, log_like, log_priors, rejected = \
                self._perform_mcmc_step(inputs, cov, log_like, log_priors, 1)
            chain[:, :, i] = inputs

            cov = self.adapt_proposal_cov(cov, chain, i, adapt_interval,
                                          adapt_delay)
        return chain

    @abstractmethod
    def evaluate_model(self, inputs):
        '''
        Calls self._model with "inputs" and returns corresponding outputs.
        '''

    @rank_zero_output_only
    def sample_from_priors(self, num_samples):
        samples = []
        for i, p in enumerate(self._priors):
            samples.append(p.rvs(num_samples).T)
        return np.vstack(samples).T

    def evaluate_log_priors(self, inputs):
        prior_dims = self._get_prior_dims()

        if inputs.shape[1] != sum(prior_dims):
            raise ValueError("Num prior distributions != num input params")

        prior_probs = np.empty((inputs.shape[0], len(self._priors)))
        in_start_idx = 0
        for i, p in enumerate(self._priors):
            in_ = inputs[:, in_start_idx:in_start_idx + prior_dims[i]]
            prior_probs[:, i] = p.pdf(in_).squeeze()
            in_start_idx += prior_dims[i]

        nonzero_priors = prior_probs != 0
        log_priors = np.full(prior_probs.shape, -np.inf)
        log_priors = np.log(prior_probs, where=nonzero_priors, out=log_priors)
        return log_priors

    def _get_prior_dims(self):
        return [p.dim if hasattr(p, 'dim') else 1 for p in self._priors]

    def evaluate_log_likelihood(self, inputs):
        log_like = self._log_like_func(inputs)
        return log_like.reshape(-1, 1)

    @staticmethod
    def evaluate_log_posterior(log_likelihood, log_priors):
        return np.sum(np.hstack((log_likelihood, log_priors)), axis=1)

    @rank_zero_output_only
    def proposal(self, inputs, cov):
        chol = self._ensure_psd_cov_and_do_chol_decomp(cov)
        z = np.random.normal(0, 1, inputs.shape)
        delta = np.matmul(chol, z.T).T
        return inputs + delta

    def acceptance_ratio(self, new_log_like, old_log_like, new_log_priors,
                         old_log_priors):
        old_log_post = self.evaluate_log_posterior(old_log_like, old_log_priors)
        new_log_post = self.evaluate_log_posterior(new_log_like, new_log_priors)
        return np.exp(new_log_post - old_log_post).reshape(-1, 1)

    @staticmethod
    @rank_zero_output_only
    def get_rejections(acceptance_ratios):
        u = np.random.uniform(0, 1, acceptance_ratios.shape)
        return acceptance_ratios < u

    def adapt_proposal_cov(self, cov, chain, idx, adapt_interval, adapt_delay):

        if self._is_adapt_iteration(adapt_interval, idx, adapt_delay):
            start = self._get_window_start(idx, adapt_delay, adapt_interval)
            end = idx + 1
            n_param = chain.shape[1]
            flat_chain = [chain[:, i, start:end].flatten() \
                          for i in range(n_param)]
            return np.cov(flat_chain)
        return cov

    def _initialize_probabilities(self, inputs):
        log_priors = self.evaluate_log_priors(inputs)
        self._check_log_priors_for_zero_probability(log_priors)
        log_like = self.evaluate_log_likelihood(inputs)
        return log_priors, log_like

    def _perform_mcmc_step(self, inputs, cov, log_like, log_priors, phi):
        new_inputs = self.proposal(inputs, cov)
        new_log_priors = self.evaluate_log_priors(new_inputs)
        new_log_like = self._eval_log_like_if_prior_nonzero(new_log_priors,
                                                            new_inputs)

        accpt_ratio = self.acceptance_ratio(new_log_like * phi,
                                            log_like * phi,
                                            new_log_priors, log_priors)

        rejected = self.get_rejections(accpt_ratio)

        inputs = np.where(rejected, inputs, new_inputs)
        log_like = np.where(rejected, log_like, new_log_like)
        log_priors = np.where(rejected, log_priors, new_log_priors)

        return inputs, log_like, log_priors, rejected

    @staticmethod
    def _is_adapt_iteration(adapt_interval, idx, adapt_delay):
        if adapt_interval is None:
            return False
        surpassed_delay = idx >= adapt_delay
        is_adapt_iteration = (idx - adapt_delay) % adapt_interval == 0
        return surpassed_delay and is_adapt_iteration

    @staticmethod
    def _get_window_start(idx, adapt_delay, adapt_interval):
        if idx >= adapt_delay + adapt_interval:
            return adapt_delay + 1
        return max(adapt_delay - adapt_interval + 1, 1)

    def _check_log_priors_for_zero_probability(self, log_priors):
        if any(~self._row_has_nonzero_prior_probability(log_priors)):
            raise ValueError('Initial inputs are out of bounds; '
                             f'prior log prob = {log_priors}')

    def _eval_log_like_if_prior_nonzero(self, log_priors, inputs):
        pos_rows = self._row_has_nonzero_prior_probability(log_priors)
        log_likes = np.zeros((log_priors.shape[0], 1))
        if inputs[pos_rows].size != 0:
            log_likes[pos_rows] = self.evaluate_log_likelihood(inputs[pos_rows])
        return log_likes

    @staticmethod
    def _row_has_nonzero_prior_probability(log_priors):
        return ~(log_priors == -np.inf).any(axis=1)

    @staticmethod
    def _ensure_psd_cov_and_do_chol_decomp(cov):
        '''
        Higham NJ. Computing a nearest symmetric positive semidefinite matrix.
        Linear Algebra and its Applications. 1988 May;103(C):103-118.

        Code implementation: https://stackoverflow.com/a/63131250/4733085
        '''
        try:
            return np.linalg.cholesky(cov)
        except:
            warnings.warn('Covariance matrix is not positive semi-definite; '
                          'forcing negative eigenvalues to zero and rebuilding '
                          'covariance matrix.')
            eigval, eigvec = np.linalg.eigh(cov)
            eigval[eigval < 0] = 0
            cov = np.dot(eigvec, np.diag(eigval)).dot(eigvec.T)
            return np.linalg.cholesky(cov)



'''
Notices:
Copyright 2018 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRessED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNess FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLess THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.
'''

import numpy as np

from abc import ABC, abstractmethod

from .smc.initializer import Initializer
from .smc.mutator import Mutator
from .utils.storage import InMemoryStorage
from .utils.mpi_utils import rank_zero_run_only
from .utils.context_manager import ContextManager


class SamplerBase:

    def __init__(self, mcmc_kernel):
        self._mcmc_kernel = mcmc_kernel
        self._initializer = Initializer(self._mcmc_kernel)
        self._mutator = Mutator(self._mcmc_kernel)
        self._updater = None
        self._mutation_ratio = 1
        self._step_list = []
        self._phi_sequence = []

        try:
            self._result = ContextManager.get_context()
        except:
            self._result = InMemoryStorage()

    @property
    def phi_sequence(self):
        return np.array(self._phi_sequence)

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, step):
        if step is not None:
            self._save_step(step)
            self._step = step

    @abstractmethod
    def sample(self):
        '''
        Performs SMC sampling. Returns step list and estimates of marginal
        log likelihood at each step.
        '''
        raise NotImplementedError

    def _initialize(self, num_particles, proposal):
        if self._result and self._result.is_restart:
            self._step = self._result[-1]
            self._phi_sequence = self._result.phi_sequence
            return None
        elif proposal:
            return self._initializer.init_particles_from_samples(*proposal)
        return self._initializer.init_particles_from_prior(num_particles)

    def _do_smc_step(self, phi, delta_phi, num_mcmc_samples):
        particles = self._updater.update(self.step, delta_phi)
        mut_particles = self._mutator.mutate(particles, phi, num_mcmc_samples)
        self._compute_mutation_ratio(particles, mut_particles)
        self.step = mut_particles

    def _compute_mutation_ratio(self, old_particles, new_particles):
        mutated = ~np.all(new_particles.params == old_particles.params, axis=1)
        self._mutation_ratio = sum(mutated) / new_particles.params.shape[0]

    @rank_zero_run_only
    def _save_step(self, step):
        self._result.save_step(step)



import abc

# compatible with Python 2 *and* 3:
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

class MCMCKernel(ABC):

    def __init__(self):
        pass

    @abc.abstractmethod
    def mutate_particles(self, particles, num_samples, proposal_cov, phi):
        return new_param_values, new_log_likelihoods

    @abc.abstractmethod
    def sample_from_prior(self, num_samples):
        return param_values

    @abc.abstractmethod
    def get_log_likelihoods(self, param_dict):
        return log_likelihoods

    @abc.abstractmethod
    def get_log_priors(self, param_dict):
        return log_priors



import numpy as np


class Checks(object):

    @classmethod
    def _is_dict(class_, input_):
        return isinstance(input_, dict)

    @classmethod
    def _is_1D_array(class_, input_):
        return input_.ndim != 1

    @classmethod
    def _is_integer_or_float(class_, input_):
        return any((class_._is_integer(input_), class_._is_float(input_)))

    @classmethod
    def _is_string_or_none(class_, input_):
        return any((class_._is_string(input_), class_._is_none(input_)))

    @staticmethod
    def _is_integer(input_):
        return isinstance(input_, int)

    @staticmethod
    def _is_float(input_):
        return isinstance(input_, float)

    @staticmethod
    def _is_string(input_):
        return isinstance(input_, str)

    @staticmethod
    def _is_none(input_):
        return input_ is None

    @staticmethod
    def _is_negative(input_):
        return input_ < 0

    @staticmethod
    def _is_positive(input_):
        return input_ > 0

    @staticmethod
    def _is_zero(input_):
        return input_ == 0

    @staticmethod
    def _is_positive_definite(input_):
        try:
            np.linalg.cholesky(input_)
            return True
        except np.linalg.linalg.LinAlgError:
            return False

    @staticmethod
    def _raise_type_error(input_, type_):
        raise TypeError('%s must be %s.' % (input_, type_))

    @staticmethod
    def _raise_negative_error(input_):
        raise ValueError('%s must be > 0.' % input_)

    @staticmethod
    def _raise_zero_error(input_):
        raise ValueError('%s cannot be zero.' % input_)

    @staticmethod
    def _raise_out_of_bounds_error(input_):
        raise ValueError('%s cannot be greater than num_time_steps.' % input)



import numpy as np

from .mcmc_base import MCMCBase
from ..log_likelihoods import Normal


class VectorMCMC(MCMCBase):
    
    def __init__(self, model, data, priors, log_like_args=None,
                 log_like_func=Normal):
        '''
        :param model: maps inputs to outputs
        :type model: callable
        :param data: data corresponding to model outputs
        :type data: 1D array
        :param priors: random variable objects with a pdf and rvs method (e.g.
            scipy stats random variable objects)
        :type priors: list of objects
        :param log_like_args: any fixed parameters that define the likelihood
            function (e.g., standard deviation for a Gaussian likelihood).
        :type log_like_args: 1D array or None
        :param log_like_func: log likelihood function that takes inputs, model,
            data, and hyperparameters and returns log likelihoods
        :type log_like_func: callable
        '''
        super().__init__(model, data, priors, log_like_args, log_like_func)

    def evaluate_model(self, inputs):
        return self._eval_model(inputs)



from checks import Checks
from ..hdf5.hdf5_storage import HDF5Storage
from ..smc.smc_step import SMCStep


class Properties(Checks):

    def __init__(self):
        super(Properties, self).__init__()
        self._num_particles = 1
        self._num_time_steps = 1
        self._temp_schedule = [0.0, 1.0]
        self._num_mcmc_steps = 1
        self._ess_threshold = 0
        self._autosaver = None
        self._restart_time_step = 0
        self._particle_chain = SMCStep()

    @property
    def num_particles(self):
        return self._num_particles

    @num_particles.setter
    def num_particles(self, num_particles):
        input_ = 'num_particles'
        if not self._is_integer(num_particles):
            self._raise_type_error(input_, 'integer')
        if self._is_negative(num_particles):
            self._raise_negative_error(input_)
        if self._is_zero(num_particles):
            self._raise_zero_error(input_)
        self._num_particles = num_particles
        return None

    @property
    def num_time_steps(self):
        return self._num_time_steps

    @num_time_steps.setter
    def num_time_steps(self, num_time_steps):
        input_ = 'num_time_steps'
        if not self._is_integer(num_time_steps):
            self._raise_type_error(input_, 'integer')
        if self._is_negative(num_time_steps):
            self._raise_negative_error(input_)
        if self._is_zero(num_time_steps):
            self._raise_zero_error(input_)
        self._num_time_steps = num_time_steps
        return None

    @property
    def temp_schedule(self):
        return self._temp_schedule

    @temp_schedule.setter
    def temp_schedule(self, temp_schedule):
        self._temp_schedule = temp_schedule

    @property
    def num_mcmc_steps(self):
        return self._num_mcmc_steps

    @num_mcmc_steps.setter
    def num_mcmc_steps(self, num_mcmc_steps):
        input_ = 'num_mcmc_steps'
        if not self._is_integer(num_mcmc_steps):
            self._raise_type_error(input_, 'integer')
        if self._is_negative(num_mcmc_steps):
            self._raise_negative_error(input_)
        if self._is_zero(num_mcmc_steps):
            self._raise_zero_error(input_)
        self._num_mcmc_steps = num_mcmc_steps
        return None

    @property
    def ess_threshold(self):
        return self._ess_threshold

    @ess_threshold.setter
    def ess_threshold(self, ess_threshold):
        input_ = 'ess_threshold'
        if ess_threshold is None:
            ess_threshold = 0.5 * self.num_particles
        if not self._is_integer_or_float(ess_threshold):
            self._raise_type_error(input_, 'float or int')
        if self._is_negative(ess_threshold):
            self._raise_negative_error(input_)
        self._ess_threshold = ess_threshold
        return None

    @property
    def autosaver(self):
        return self._autosaver

    @autosaver.setter
    def autosaver(self, autosave_file):
        if not self._is_string_or_none(autosave_file):
            self._raise_type_error('autosave_file', 'string or None')
        if self._rank == 0 and autosave_file is not None:
            self._autosaver = HDF5Storage(autosave_file, mode='w')
        else:
            self._autosaver = None
        return None

    @property
    def restart_time_step(self):
        return self._restart_time_step

    @restart_time_step.setter
    def restart_time_step(self, restart_time_step):
        input_ = 'restart_time_step'
        if not self._is_integer(restart_time_step):
            self._raise_type_error(input_, 'integer')
        if self._is_negative(restart_time_step):
            self._raise_negative_error(input_)
        if restart_time_step > self.num_time_steps:
            self._raise_out_of_bounds_error(input_)
        self._restart_time_step = restart_time_step
        return None

    @property
    def particle_chain(self):
        return self._particle_chain

    @particle_chain.setter
    def particle_chain(self, particle_chain):
        input_ = 'particle_chain'
        if not isinstance(particle_chain, SMCStep) and \
           not self._is_none(particle_chain):
            self._raise_type_error(input_, 'SMCStep instance or None')
        self._particle_chain = particle_chain
        return None



'''
Notices:
Copyright 2018 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRessED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNess FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLess THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.
'''

import imp
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import seaborn as sns

from matplotlib.backends.backend_pdf import PdfPages
from scipy.stats import gaussian_kde


def _mpi_decorator(func):
    def wrapper(self, *args, **kwargs):
        """
        Detects whether multiple processors are available and sets
        self.number_CPUs and self.cpu_rank accordingly. Only calls decorated
        function using rank 0.
        """
        try:
            imp.find_module('mpi4py')

            from mpi4py import MPI
            comm = MPI.COMM_WORLD.Clone()

            size = comm.size
            rank = comm.rank
            comm = comm

        except ImportError:

            size = 1
            rank = 0
            comm = SingleRankComm()

        if rank == 0:
            func(self, *args, **kwargs)

    return wrapper


def plot_mcmc_chain(chain, param_labels, burnin=0, save=False, show=True,
                    include_kde=False, filename='chain.png',
                    report_style=False):

    n_columns = 2
    gridspec = {'width_ratios': [1., 0.], 'wspace': 0.0}
    if include_kde:
        gridspec = {'width_ratios': [0.85, 0.15], 'wspace': 0.0}

    fig, ax = plt.subplots(len(param_labels), n_columns, sharey='row',
                           gridspec_kw=gridspec)

    chain = chain[:, :, burnin:]
    for i, name in enumerate(param_labels):
        for parallel_chain in chain:
            ax[i, 0].plot(parallel_chain[i], '-', linewidth=0.5)

            if include_kde:
                ylims = ax[i, 0].get_ylim()
                x = np.linspace(ylims[0], ylims[1], 1000)
                kde = gaussian_kde(parallel_chain[i])
                ax[i, 1].plot(kde.pdf(x), x, '-')
                ax[i, 1].fill_betweenx(x, kde.pdf(x), np.zeros(x.shape),
                                      alpha=0.3)

        ax[i, 1].axis('off')
        if include_kde:
            ax[i, 1].set_xlim(0, None)

        ax[i, 0].set_ylabel(name)
        ax[i, 0].set_xlim(0, chain.shape[2]) 
        ax[i, 0].get_xaxis().set_visible(False)

    ax[len(param_labels) - 1, 0].get_xaxis().set_visible(True)
    ax[len(param_labels) - 1, 0].set_xlabel('sample #')
    ax[len(param_labels) - 1, 1].set_xlabel('probability density')

    if save:
        plt.tight_layout()
        plt.savefig(filename, dpi=150)
    if show:
        plt.show()

    if report_style:
        fig.set_figheight(fig.get_figheight() * ax.shape[0])
        rprt_filename = os.path.splitext(filename)[0] + '_report.pdf'
        with PdfPages(rprt_filename) as pdf:
            scale = fig.dpi_scale_trans.inverted()
            extent1 = ax[0, 0].get_window_extent().transformed(scale)
            extent2 = ax[1, 0].get_window_extent().transformed(scale)
            ax_height = extent1.y1 - extent2.y1

            for i in range(ax.shape[0]):
                ax[i, 0].get_xaxis().set_visible(True)
                ax[i, 0].set_xlabel('sample #')
                ax[i, 1].set_xlabel('probability density')
                extent = ax[i, 0].get_window_extent().transformed(scale)
                extent.x0 = 0
                extent.x1 = fig.get_figwidth()
                extent.y0 = extent.y1 - ax_height
                pdf.savefig(plt.gcf(), bbox_inches=extent)

    return fig

def plot_pairwise(samples, weights=None, param_names=None,
                  save=False, show=True, xlim=None, ylim=None,
                  filename='pairwise.png'):  # pragma no cover
    #TODO true params on plot #true_params=None,
    '''
    Plots pairwise distributions of all parameter combos. Color codes each
    by weight if provided.
    '''
    if param_names is None:
        param_names = [f'p{i}' for i in range(samples.shape[1])]

    if weights is None:
        weights = np.ones((samples.shape[0], 1))

    columns = param_names + ['weights_']
    samples = np.hstack((samples, weights))

    sm = plt.cm.ScalarMappable(cmap='viridis')
    sm.set_array([])

    df = pd.DataFrame(samples, columns=columns)

    ax = sns.pairplot(df, diag_kind='kde', corner=True,
                      hue='weights_', palette='viridis',
                      diag_kws={'weights': weights.flatten(), 'hue': None})
    ax.set(xlim=xlim)
    ax.set(ylim=ylim)

    ax.legend.remove()
    cbar = plt.gcf().colorbar(sm)
    cbar.ax.set_ylabel('Particle Weight')

    # compute means TODO plot means
    if weights is None:
        means = np.mean(samples, axis=0)
    else:
        means = np.sum(samples * weights, axis=0)

    if save:
        plt.savefig(filename)
    if show:
        plt.show()

    return ax


def plot_geweke(burnin, z, param_labels=None):

    n_params = z[0].shape[0]
    if param_labels is None:
        param_labels = [f"Param{i}" for i in range(n_params)]

    xlim = (0, int(burnin[-1] * 1.1))
    y = np.ones(2)

    fig, ax = plt.subplots(n_params)
    for i in range(n_params):
        ax[i].fill_between([0, xlim[1]], y*2, y*-2, alpha=0.5, color="0.4")
        ax[i].fill_between([0, xlim[1]], y*1, y*-1, alpha=0.5, color="0.4")
        ax[i].axhline(2, linestyle="--", color="k")
        ax[i].axhline(-2, linestyle="--", color="k")
        ax[i].plot(burnin, z[:, i], "o")
        ax[i].set_xlim(xlim)
        ax[i].set_ylabel(param_labels[i])

    ax[0].set_title("Geweke Scores")
    ax[-1].set_xlabel("Burnin")
    plt.show()



'''
Notices:
Copyright 2018 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRessED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNess FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLess THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.
'''

import numpy as np

from scipy.optimize import bisect
from tqdm import tqdm

from .sampler_base import SamplerBase
from .smc.updater import Updater
from .utils.progress_bar import set_bar
from .utils.mpi_utils import rank_zero_output_only


class FixedSampler(SamplerBase):
    '''
    SMC sampler using a fixed phi sequence.
    '''

    def __init__(self, mcmc_kernel):
        '''
        :param mcmc_kernel: a kernel object for conducting particle mutation
        :type mcmc_kernel: MCMCKernel object
        '''
        super().__init__(mcmc_kernel)

    def sample(self, num_particles, num_mcmc_samples, phi_sequence,
               ess_threshold, proposal=None, progress_bar=True):
        '''
        :param num_particles: number of particles
        :type num_particles: int
        :param num_mcmc_samples: number of MCMC samples to draw from the
            MCMC kernel per iteration per particle
        :type num_mcmc_samples: int
        :param phi_sequence: increasing monotonic sequence of floats starting
            at 0 and ending at 1; sometimes referred to as tempering schedule
        :type phi_sequence: list or array
        :param ess_threshold: the effective sample size at which resampling
            should be conducted; given as a fraction of num_particles and must
            be in the range [0, 1]
        :type ess_threshold: float
        :param proposal: tuple of samples from a proposal distribution used to
            initialize the SMC sampler; first element is a dictionary with keys
            equal to parameter names and values equal to corresponding samples;
            second element is array of corresponding proposal PDF values
        :type proposal: tuple(dict, array)
        :param progress_bar: display progress bar during sampling
        :type progress_bar: bool
        '''
        self._updater = Updater(ess_threshold)
        self._phi_sequence = phi_sequence

        self.step = self._initialize(num_particles, proposal)

        phi_iterator = self._phi_sequence[1:]
        if progress_bar:
            phi_iterator = tqdm(phi_iterator)
        set_bar(phi_iterator, 1, self._mutation_ratio, self._updater)

        for i, phi in enumerate(phi_iterator):
            dphi = phi - self._phi_sequence[i]
            self._do_smc_step(phi, dphi, num_mcmc_samples)
            set_bar(phi_iterator, i + 2, self._mutation_ratio, self._updater)

        return self._result, self._result.estimate_marginal_log_likelihoods()



class AdaptiveSampler(SamplerBase):
    '''
    SMC sampler using an adaptive phi sequence.
    '''
    def __init__(self, mcmc_kernel):
        '''
        :param mcmc_kernel: a kernel object for conducting particle mutation
        :type mcmc_kernel: MCMCKernel object
        '''
        self.req_phi_index = None
        super().__init__(mcmc_kernel)

    def sample(self, num_particles, num_mcmc_samples, target_ess=0.8,
               proposal=None, required_phi=1, progress_bar=True):
        '''
        :param num_particles: number of particles
        :type num_particles: int
        :param num_mcmc_samples: number of MCMC samples to draw from the
            MCMC kernel per iteration per particle
        :type num_mcmc_samples: int
        :param target_ess: controls adaptive stepping by picking the next
            phi such that the effective sample size is equal to the threshold.
            Specified as a fraction of total number particles (between 0 and 1).
        :type target_ess: float
        :param proposal: tuple of samples from a proposal distribution used to
            initialize the SMC sampler; first element is a dictionary with keys
            equal to parameter names and values equal to corresponding samples;
            second element is array of corresponding proposal PDF values
        :type proposal: tuple(dict, array)
        :param required_phi: specific values of phi that must be included in
            the phi sequence (regardless of optimized step)
        :type required_phi: float, int, or list
        '''
        self._updater = Updater(ess_threshold=1) # ensure always resampling
        self._phi_sequence = [0]

        self.step = self._initialize(num_particles, proposal)

        pbar = self._init_progress_bar(progress_bar)

        while self._phi_sequence[-1] < 1:
            phi = self.optimize_step(self.step, self._phi_sequence[-1],
                                     target_ess, required_phi)
            dphi = phi - self._phi_sequence[-1]
            self._do_smc_step(phi, dphi, num_mcmc_samples)
            self._phi_sequence.append(phi)
            self._update_progress_bar(pbar, dphi)

        self._close_progress_bar(pbar)

        self.req_phi_index = [i for i, phi in enumerate(self.phi_sequence) \
                              if phi in self._as_phi_list(required_phi)]

        return self._result, self._result.estimate_marginal_log_likelihoods()

    @rank_zero_output_only
    def optimize_step(self, particles, phi_old, target_ess=1, required_phi=1):
        phi = 1
        if not self._full_step_meets_target(phi_old, particles, target_ess):
            phi = bisect(self.predict_ess_margin, phi_old, 1,
                         args=(phi_old, particles, target_ess))
        proposed_phi_list = self._as_phi_list(required_phi)
        proposed_phi_list.append(phi)
        return self._select_phi(proposed_phi_list, phi_old)

    def predict_ess_margin(self, phi_new, phi_old, particles, target_ess=1):
        beta = np.exp((phi_new - phi_old) * particles.log_likes)
        numer = np.sum(beta) ** 2
        denom = np.sum(beta ** 2)
        ESS = 0
        if numer > 0 and denom > 0:
            ESS = numer / denom
        return ESS - particles.num_particles * target_ess

    @staticmethod
    def _as_phi_list(phi):
        if not isinstance(phi, list):
            phi = [phi]
        phi = sorted([p for p in phi if p < 1])
        return phi

    @staticmethod
    def _select_phi(proposed_phi_list, phi_old):
        return min([p for p in proposed_phi_list if p > phi_old])

    def _full_step_meets_target(self, phi_old, particles, target_ess):
        return self.predict_ess_margin(1, phi_old, particles, target_ess) > 0

    def _init_progress_bar(self, progress_bar):
        pbar = False
        if progress_bar:
            bar_format =  "{desc}: {percentage:.2f}%|{bar}| " + \
                       "phi: {n:.5f}/{total_fmt} [{elapsed}<{remaining}"
            pbar = tqdm(total=1.0, bar_format = bar_format)
            pbar.set_description('[ mutation ratio: n/a')
            pbar.update(0)
        return pbar

    def _update_progress_bar(self, pbar, dphi):
        if pbar:
            pbar.set_description(f'[ mutation ratio: {self._mutation_ratio}')
            pbar.update(dphi)

    @staticmethod
    def _close_progress_bar(pbar):
        if pbar:
            pbar.close()



import threading

class ContextManager:
    _contexts = threading.local()

    def __enter__(self):
        type(self).get_contexts().append(self)
        return self

    def __exit__(self, _type, _value, _traceback):
        type(self).get_contexts().pop()

    @classmethod
    def get_contexts(cls):
        """Returns context stack"""
        if not hasattr(cls._contexts, 'stack'):
            cls._contexts.stack = []
        return cls._contexts.stack

    @classmethod
    def get_context(cls):
        """Return the deepest context on the stack."""
        return cls.get_contexts()[-1]



class SingleRankComm():

    def __init__(self):
        pass

    @staticmethod
    def Get_rank():
        return 0

    @staticmethod
    def Get_size():
        return 1

    @staticmethod
    def scatter(scatter_list, *args, **kwargs):
        if len(scatter_list) == 1:
            return scatter_list[0]
        else:
            raise ValueError("SingleRankComm cannot accept more than one " +\
                             "partition")

    @staticmethod
    def allgather(scatter_list, *args, **kwargs):
        return [scatter_list]



import numpy as np

from scipy.signal import welch


def compute_geweke(x, window_pct=10, step_pct=1):
    '''
    :param x: MCMC chain with shape (number samples, number parameters)
    :type x: 2D array
    :param window_pct: percentage of total samples to use when computing
        the Geweke score
    :type window_pct: int or float
    :param step_pct: window shift for each Geweke score calculation given
        as a percentage of the total samples
    :type step_pct: int or float
    '''
    if window_pct > 50 or window_pct <= 0:
        raise ValueError("window size must be in [0, 50] percent")

    n_samples = x.shape[1]

    burnins_to_test = np.arange(0, 50 - window_pct + step_pct, step_pct)
    burnins_to_test = burnins_to_test * n_samples / 100

    window = int(window_pct * n_samples / 100)

    x2 = x[:, int(0.5 * n_samples):]
    psd2 = _spec_density(x2)
    mean2 = x2.mean(axis=1)

    z = []
    burnin = []
    for burn in burnins_to_test:

        burnin.append(int(burn))

        x1 = x[:, burnin[-1]:burnin[-1] + window]
        psd1 = _spec_density(x1)
        mean1 = x1.mean(axis=1)

        z.append((mean1 - mean2) / \
                 np.sqrt(psd1 / x1.shape[1] + psd2 / x2.shape[1]))

    return burnin, np.array(z)


def _spec_density(x):
    psd = welch(x, detrend=False, axis=1, scaling='spectrum')[1][:, 0]
    return psd



'''
Notices:
Copyright 2018 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRessED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNess FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLess THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.
'''

import numpy as np

from copy import copy

from .particles import Particles
from ..mcmc.kernel_base import MCMCKernel
from ..utils.single_rank_comm import SingleRankComm


class Initializer:
    '''
    Generates SMCStep objects based on either samples from a prior distribution
    or given input samples from a sampling distribution.
    '''
    def __init__(self, mcmc_kernel):
        '''
        :param mcmc_kernel: a kernel object for conducting particle mutation
        :type mcmc_kernel: MCMCKernel object
        '''
        self.mcmc_kernel = mcmc_kernel

    @property
    def mcmc_kernel(self):
        return self._mcmc_kernel

    @mcmc_kernel.setter
    def mcmc_kernel(self, mcmc_kernel):
        if not isinstance(mcmc_kernel, MCMCKernel):
            raise TypeError
        self._mcmc_kernel = mcmc_kernel

    def init_particles_from_prior(self, num_particles):
        '''
        Use model stored in MCMC kernel to sample initial set of particles.

        :param num_particles: number of particles to sample (total across all
            ranks)
        :type num_particles: int
        '''
        params = self.mcmc_kernel.sample_from_prior(num_particles)
        log_likes = self.mcmc_kernel.get_log_likelihoods(params)
        log_weights = np.array([np.log(1/num_particles)] * num_particles)
        particles = Particles(params, log_likes, log_weights)
        particles.attrs.update({'phi': 0})
        return particles

    def init_particles_from_samples(self, samples, proposal_pdensities):
        '''
        Initialize a set of particles using pre-sampled parameter values and
        the corresponding prior pdf at those values.

        :param samples: samples of parameters used to initialize particles;
            must be a dictionary with keys = parameter names and values =
            parameter values. Can also be a pandas DataFrame object for this
            reason.
        :type samples: dict, pandas.DataFrame
        :param proposal_pdensities: corresponding probability density function
            values; must be aligned with samples
        :type proposal_pdensities: list or nd.array
        '''
        proposal_pdensities = np.array(proposal_pdensities).reshape(-1, 1)
        log_likes = self.mcmc_kernel.get_log_likelihoods(samples)
        log_priors = self.mcmc_kernel.get_log_priors(samples)
        log_weights = log_priors - np.log(proposal_pdensities)
        particles = Particles(samples, log_likes, log_weights)
        particles.attrs.update({'phi': 0})
        return particles



from copy import copy
import numpy as np

from .kernel_base import MCMCKernel

class VectorMCMCKernel(MCMCKernel):

    def __init__(self, vector_mcmc_object, param_order):
        self._mcmc = vector_mcmc_object
        self._param_order = param_order

    def mutate_particles(self, param_dict, log_likes, num_samples, cov, phi):
        param_array = self._conv_param_dict_to_array(param_dict)
        param_array, log_likes = self._mcmc.smc_metropolis(param_array,
                                                           num_samples,
                                                           cov, phi)
        param_dict = self._conv_param_array_to_dict(param_array)
        return param_dict, log_likes

    def sample_from_prior(self, num_samples):
        param_array = self._mcmc.sample_from_priors(num_samples)
        return self._conv_param_array_to_dict(param_array)

    def get_log_likelihoods(self, param_dict):
        param_array = self._conv_param_dict_to_array(param_dict)
        return self._mcmc.evaluate_log_likelihood(param_array)

    def get_log_priors(self, param_dict):
        param_array = self._conv_param_dict_to_array(param_dict)
        log_priors = self._mcmc.evaluate_log_priors(param_array)
        return np.sum(log_priors, axis=1).reshape(-1, 1)

    def _conv_param_array_to_dict(self, param_array):
        return dict(zip(self._param_order, param_array.T))

    def _conv_param_dict_to_array(self, param_dict):
        dim0 = 1
        if not isinstance(param_dict[self._param_order[0]], (int, float)):
            dim0 = len(param_dict[self._param_order[0]])
        param_array = np.zeros((dim0, len(self._param_order)))

        for i, k in enumerate(self._param_order):
            param_array[:, i] = param_dict[k]
        return param_array



import numpy as np

class BaseLogLike:

    def __init__(self, model, data, args):
        self._model = model
        self._data = data
        self._args = args

        self.set_model_wrapper(lambda model, x: model(x))

    def _get_output(self, inputs):
        output = self._model_wrapper(self._model, inputs)
        if np.isnan(output).any():
            raise ValueError
        return output

    def set_model_wrapper(self, wrapper):
        self._model_wrapper = wrapper



class Normal(BaseLogLike):

    def __init__(self, model, data, args):
        '''
        Likelihood function for data with additive, iid errors sampled from a
        normal distribution with mean = 0 and std_dev = args. If args is None,
        assumes that the last column of inputs contains the std_dev value.
        '''
        super().__init__(model, data, args)

    def __call__(self, inputs):
        std_dev = self._args
        if std_dev is None:
            std_dev = inputs[:, -1]
            inputs = inputs[:, :-1]
        var = std_dev ** 2

        output = self._get_output(inputs)

        return self._calc_normal_log_like(output, self._data, var)

    @staticmethod
    def _calc_normal_log_like(output, data, var):
        ssqe = np.sum((output - data) ** 2, axis=1)
    
        term1 = -np.log(2 * np.pi * var) * (output.shape[1] / 2.) 
        term2 = -1 / 2. * ssqe / var
    
        return (term1 + term2)


class MultiSourceNormal(Normal):

    def __init__(self, model, data, args):
        '''
        Likelihood function for data with additive, iid errors sampled from N 
        normal distributions, each with mean = 0 and std_dev = args[1][n] for
        n = 1, ..., N. The number of data points sampled from each respective
        distribution should be provided in args[1]. Both args[0] and args[1]
        should be aligned and correspond to the ordering of data. That is, if
        args[0] = (2, 3, 1), then len(data) = 6 and the first two data points
        are assumed to be drawn from a Normal distribution w/
        std_dev = args[1][0]. If args[1][n] is None, assumes that the last M
        columns of inputs contains the std_dev values, where M is the total
        number of Nones in args[1]. The std_dev columns should be aligned with
        args[0] in this case.

        :param args: data segment lengths and corresponding standard deviations
            N from Gaussian distribution
        :type args: list of two tuples, each of length N
        '''
        super().__init__(model, data, args)
        self._num_nones = self._args[1].count(None)

        if sum(args[0]) != data.shape[0]:
            raise ValueError("data segments in args[0] must sum to dim of data")

    def __call__(self, inputs):

        std_devs, inputs = self._process_fixed_and_variable_std(inputs)

        output = self._get_output(inputs)

        log_likes = []
        start_idx = 0
        for segment_len, segment_std_dev in zip(*[self._args[0], std_devs]):

            if segment_len == 0:
                continue

            output_segment = output[:, start_idx : start_idx + segment_len]
            data_segment = self._data[start_idx : start_idx + segment_len]
            log_likes.append(self._calc_normal_log_like(output_segment,
                                                        data_segment,
                                                        segment_std_dev ** 2))
            start_idx += segment_len

        return np.sum(log_likes, axis=0)


    def _process_fixed_and_variable_std(self, inputs):
        '''
        Identifies standard deviations to be estimated and pulls appropriate
        samples from the input array.
        '''
        std_devs = self._args[1]
        new_std_devs = []
        j = 0
        for i, std in enumerate(std_devs):
            if std is None:
                new_std_devs.append(inputs[:, -self._num_nones + j])
                j += 1
            else:
                new_std_devs.append(std)
    
        new_inputs = inputs.copy()
        if self._num_nones > 0:
            new_inputs = new_inputs[:, :-self._num_nones]
    
        return tuple(new_std_devs), new_inputs


class MVNormal(BaseLogLike):

    def __init__(self, model, data, args):
        '''
        Likelihood function for data with additive, iid errors sampled from N-D
        multivariate normal distribution with mean = [0] * N and covariances = 
        args[n] for n = 1, ..., N where, for example, a 4-D covariance
        matrix is defined as:

                      [0 1 2 3]
                cov = [1 4 5 6]  where 0, ..., 9 represent entries in args
                      [2 5 7 8]
                      [3 6 8 9]
        
        If args[n] is None, assumes that the last M columns of inputs
        contains covariance samples, where M is the total number of Nones in
        args (Nones do not have to be contiguous).

        Data array should have shape (number of snapshots, number of features)
        where snapshots are independent simultaneous measurements of all the
        data features. Features, for example, might be sensor readings at
        different locations. Model should output array with shape (number of
        samples/particles, number of features).

        :param args: covariances of the N-D multivariate normal
        :type args: list, len(args) = N * (N + 1) / 2
        '''
        super().__init__(model, data, args)
        self._num_nones = self._args.count(None)

    def __call__(self, inputs):
        cov_arg_array, inputs = self._process_fixed_and_variable_covar(inputs)
        cov_matrices = np.tile(self._get_cov(cov_arg_array),
                               (self._data.shape[0], 1, 1, 1))

        data = np.expand_dims(self._data, 1)
        output = self._get_output(inputs)
        error = output - data
        error = np.expand_dims(error, 2)
        errorT = np.transpose(error, axes=(0, 1, 3, 2))

        term1 = - self._data.shape[1] / 2 * np.log(2 * np.pi)
        term2 = - 1 / 2 * np.log(np.linalg.det(cov_matrices))
        term2 = np.expand_dims(term2, (2, 3))
        term3 = np.matmul(np.matmul(error, np.linalg.inv(cov_matrices)), errorT)

        log_likes = term1 + term2 + -(1 / 2) * term3
        log_likes = np.sum(log_likes, axis=0)

        return log_likes[:, :, 0]

    def _get_cov(self, cov_args):
        d = int((np.sqrt(1 + 8 * cov_args.shape[1]) - 1) / 2)
        covs = np.zeros((cov_args.shape[0], d, d))
        p, q = np.triu_indices(d)
        covs[:, p, q] = cov_args
        covs += np.transpose(np.triu(covs, 1), axes=[0, 2, 1])
        return covs

    def _process_fixed_and_variable_covar(self, inputs):
        covars = np.tile(self._args, (inputs.shape[0], 1))
        j = 0
        for i, arg in enumerate(self._args):
            if arg is None:
                covars[:, i] = inputs[:, -self._num_nones + j]
                j += 1

        new_inputs = inputs.copy()
        if self._num_nones > 0:
            new_inputs = new_inputs[:, :-self._num_nones]
 
        return covars, new_inputs



class MVNRandomEffects(BaseLogLike):

    def __init__(self, model, data, args):
        '''
        Likelihood function for random effects.

        :param model: computational model
        :type model: callable
        :param data: data for the random effects with shape (num rand effects,
            num data features)
        :type data: 2D array
        :param args: total and random effects covariances and standard devs
        :type args: tuple (x, y)
        '''

        super().__init__(model, data, args)

        self._randeffs = None
        self._init_randeffs()

        self._n_total_eff_nones = self._args[0].count(None)
        self._n_rand_eff_nones = self._args[1].count(None)
        self._n_nones = self._n_total_eff_nones + self._n_rand_eff_nones

    def __call__(self, inputs):
        processed_inputs = self._process_fixed_and_variable_std(inputs)

        total_eff_inputs = processed_inputs[0]
        rand_eff_inputs = processed_inputs[1]
        rand_eff_model_inputs = processed_inputs[2]

        total_eff_log_like = np.full((inputs.shape[0], 1), -np.inf)
        for i in range(inputs.shape[0]):

            reff_data_array = np.array([ni[i] for ni in rand_eff_model_inputs])
            total_eff = MVNormal(self._total_effects_model, reff_data_array,
                                 self._args[0])

            in_ = total_eff_inputs[i].reshape(1, -1)
            total_eff_log_like[i, 0] = total_eff(in_)

        iterable = zip(rand_eff_inputs, self._randeffs)
        log_like = [np.c_[d(in_)] for in_, d in iterable]
        log_like.append(total_eff_log_like)

        return np.sum(log_like, axis=0)

    def set_model_wrapper(self, wrapper):
        self._model_wrapper = wrapper
        self._init_randeffs()

    def _init_randeffs(self):
        self._randeffs = []
        models = self._verify_model_type()
        for i, arg_i in enumerate(self._args[1]):
            self._randeffs.append(Normal(models[i], self._data[i], arg_i))
            self._randeffs[-1].set_model_wrapper(self._model_wrapper)

    def _verify_model_type(self):
        if isinstance(self._model, list):
            if len(self._args[1]) != len(self._model):
                raise ValueError('If multiple models provided, number must be ',
                                 'equal to number of random effects.')
            return self._model
        return [self._model] * len(self._args[1])

    def _process_fixed_and_variable_std(self, inputs):
        num_randeff = len(self._args[1])

        model_inputs = inputs
        if self._n_nones > 0:
            model_inputs = model_inputs[:, :-self._n_nones]
            like_args = inputs[:, -self._n_nones:]
            total_eff_args = like_args[:, :self._n_total_eff_nones]
            rand_eff_args = like_args[:, self._n_total_eff_nones:]

        split_inputs = np.array_split(model_inputs, num_randeff + 1, axis=1)
        total_eff_model_inputs = split_inputs[0]
        rand_eff_model_inputs = split_inputs[1:]

        total_eff_inputs = total_eff_model_inputs.copy()
        rand_eff_inputs = rand_eff_model_inputs.copy()

        if self._n_total_eff_nones > 0:
            total_eff_inputs = np.c_[total_eff_model_inputs, total_eff_args]

        if self._n_rand_eff_nones > 0:
            arg_idx = [i for i, val in enumerate(self._args[1]) if val == None]
            for i, j in enumerate(arg_idx):
                rand_eff_inputs[j] = np.c_[rand_eff_model_inputs[j],
                                           rand_eff_args[:, i]]

        return (total_eff_inputs, rand_eff_inputs, rand_eff_model_inputs)

    @staticmethod
    def _total_effects_model(inputs):
        return inputs



import h5py
import numpy as np
import os

from abc import abstractmethod

from ..smc.particles import Particles
from .context_manager import ContextManager


class BaseStorage(ContextManager):

    def __init__(self):
        self.is_restart = False

    def estimate_marginal_log_likelihoods(self):
        sum_un_log_wts = [0]
        sum_un_log_wts += [p.total_unnorm_log_weight for p in self]
        return np.cumsum(sum_un_log_wts)

    @abstractmethod
    def __getitem__(self, idx):
        raise NotImplementedError

    def __iter__(self):
        self._idx = 0
        return self

    @abstractmethod
    def __next__(self):
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def save_step(self, step):
        raise NotImplementedError



class InMemoryStorage(BaseStorage):

    def __init__(self):
        super().__init__()
        self._step_list = []
        self._phi_sequence = []

    @property
    def phi_sequence(self):
        return self._phi_sequence

    def __getitem__(self, idx):
        return self._step_list[idx]

    def __next__(self):
        if self._idx < len(self._step_list):
            output = self._step_list[self._idx]
            self._idx += 1
            return output
        else:
            raise StopIteration

    def __len__(self):
        return len(self._step_list)

    def save_step(self, step):
        self._step_list.append(step)
        self._phi_sequence.append(step.attrs['phi'])



class HDF5Storage(BaseStorage):

    def __init__(self, filename, mode='a'):
        super().__init__()
        self._filename = filename
        self._len = 0
        if os.path.exists(filename):
            if mode != 'w':
                self._init_length_on_restart()
                self.is_restart = True
            else:
                os.remove(filename)

    @property
    def phi_sequence(self):
        h5 = self._open_h5('r')
        phi_sequence = [h5[i].attrs['phi'] for i in sorted(h5.keys(), key=int)]
        self._close(h5)
        return phi_sequence

    def save_step(self, step):
        h5 = self._open_h5('a')

        step_grp = h5.create_group(str(len(self)))
        step_grp.attrs['phi'] = step.attrs['phi']
        step_grp.attrs['total_unnorm_log_weight'] = step.total_unnorm_log_weight
        step_grp.create_dataset('log_likes', data=step.log_likes)
        step_grp.create_dataset('log_weights', data=step.log_weights)

        param_grp = step_grp.create_group('params')
        for key, array in step.param_dict.items():
            param_grp.create_dataset(key, data=array)

        self._close(h5)

    def _open_h5(self, mode):
        h5 = h5py.File(self._filename, mode)
        self._len = len(h5.keys())
        return h5

    def _close(self, h5):
        self._len = len(h5.keys())
        h5.close()

    def __getitem__(self, idx):
        h5 = self._open_h5('r')
        step_grp = h5[self._format_index(idx)]

        total_unlw = step_grp.attrs['total_unnorm_log_weight']
        kwargs = {k: v[:] for k, v in step_grp.items() if k != 'params'}
        kwargs['params'] = {k: v[:] for k, v in step_grp['params'].items()}
        particles = Particles(**kwargs)
        particles._total_unlw = total_unlw

        self._close(h5)
        return particles

    def _format_index(self, idx):
        if idx < 0:
            idx = len(self) + idx
        if idx < 0 or idx >= len(self):
            raise IndexError(f'index {idx} out of range.')
        return str(idx)

    def __next__(self):
        try:
            output = self[self._idx]
            self._idx += 1
            return output
        except IndexError:
            raise StopIteration

    def __len__(self):
        return self._len

    def _init_length_on_restart(self):
        h5 = self._open_h5('r')
        self._close(h5)



def rank_zero_output_only(func):
    """
    Wrapper function that detects whether mpi4py is available and forces all
    ranks to output from rank 0 only. Intended to be used as a decorator where
    mpi-enabled MCMC kernels are deployed.
    """
    def wrapper(self, *args, **kwargs):
        output = func(self, *args, **kwargs)

        try:
            if hasattr(self, '_comm'):
                output = self._comm.bcast(output, root=0)
            else:
                output = self._mcmc_kernel._mcmc._comm.bcast(output, root=0)

        except AttributeError:
            pass

        return output

    return wrapper


def rank_zero_run_only(func):
    """
    Wrapper function that detects whether mpi4py is available and prohibits
    ranks > 0 from performing the decorated function. Only intended for
    functions that return None (eg operate on internal state).
    """
    def wrapper(self, *args, **kwargs):

        try:
            if self._mcmc_kernel._mcmc._rank == 0:
                func(self, *args, **kwargs)
                self._mcmc_kernel._mcmc._comm.Barrier()
            else:
                self._mcmc_kernel._mcmc._comm.Barrier()

        except AttributeError: # no MPI
            func(self, *args, **kwargs)

    return wrapper



from tqdm import tqdm

def set_bar(pbar, t, mutation_ratio, updater):
    if isinstance(pbar, tqdm):
        desc = "t: {:2d} | ess: {:8.2f} | mut. ratio: {:.1%} | resample: {} |"
        pbar.set_description(desc.format(t, updater.ess, mutation_ratio,
                                         updater.resampled))
    return pbar



import re
from setuptools import setup

def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), 
                       open(project + '/__init__.py').read())
    return result.group(1)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="smcpy",
    version=get_property('__version__', 'smcpy'),
    author="Patrick Leser",
    author_email="patrick.e.leser@nasa.gov",
    description="A package for performing uncertainty quantification using a parallelized sequential Monte Carlo sampler.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nasa/SMCPy",
    packages=["smcpy",
              "smcpy.smc",
              "smcpy.mcmc",
              "smcpy.utils"],
    install_requires=['numpy',
                      'h5py',
                      'matplotlib',
                      'tqdm',
                      'scipy',
                      'pandas',
                      'seaborn',
                      'pytest',
                      'pytest-mock'],
    python_requires='~=3.4',
    classifiers=[
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha"
    ],
)



import numpy as np

from scipy.stats import invwishart


class ImproperUniform:

    def __init__(self, lower_bound=None, upper_bound=None):
        self._lower = lower_bound
        self._upper = upper_bound

        if self._lower is None:
            self._lower = -np.inf

        if self._upper is None:
            self._upper = np.inf

    def pdf(self, x):
        '''
        :param x: input array
        :type x: 1D or 2D array; if 2D, must squeeze to 1D
        '''
        array_x = np.array(x).squeeze()

        if array_x.ndim > 1:
            raise ValueError('Input array must be 1D or must squeeze to 1D')

        prior_pdf = np.zeros(array_x.size)
        in_bounds = (array_x >= self._lower) & (array_x <= self._upper)
        return np.add(prior_pdf, 1, out=prior_pdf, where=in_bounds)



class InvWishart:

    def __init__(self, dof, scale):
        '''
        :param scale: scale matrix
        :param dof: degrees of freedom (if == scale.shape[0], noninformative)
        '''
        self._cov_dim = scale.shape[0]
        self._dim = int(self._cov_dim * (self._cov_dim + 1) / 2)
        self._invwishart = invwishart(dof, scale)

    @property
    def dim(self):
        return self._dim

    def rvs(self, num_samples):
        '''
        :param num_samples: number of samples to return
        :type num_samples: int
        '''
        cov = self._invwishart.rvs(num_samples)
        idx1, idx2 = np.triu_indices(self._cov_dim)
        return cov[:, idx1, idx2]

    def pdf(self, x):
        '''
        :param x: input array
        :type x: 2D array (# samples, # unique covariances)
        '''
        covs = self._assemble_covs(x)
        try:
            return self._invwishart.pdf(covs)
        except np.linalg.LinAlgError:
            return np.zeros((x.shape[0], 1))

    def _assemble_covs(self, samples):
        covs = np.zeros((samples.shape[0], self._cov_dim, self._cov_dim))
        idx1, idx2 = np.triu_indices(self._cov_dim)
        covs[:, idx1, idx2] = samples
        covs += np.transpose(np.triu(covs, 1), axes=(0, 2, 1))
        return np.transpose(covs, axes=(1, 2, 0))

class ImproperCov:
    '''
    Improper uniform prior distribution over all positive definite covariance
    matrices. If sampling using the rvs method, will return samples from an
    Inverse Wishart distribution with degrees of freedom dof and scale matrix S.
    '''
    def __init__(self, num_cov_cols, dof=None, S=None):
        self._ncols = num_cov_cols
        self._dim = int(num_cov_cols * (num_cov_cols + 1) / 2)
        self._iw_dof = dof if dof is not None else num_cov_cols
        self._iw_scale = S if S is not None else np.eye(self._ncols)

    @property
    def dim(self):
        return self._dim

    def pdf(self, samples):
        self._check_samples_shape(samples)
        covs = self._assemble_covs(samples)
        pdf = np.array([self._is_pos_semidef(x) for x in covs]).reshape(-1, 1)
        return pdf

    def rvs(self, num_samples):
        cov = invwishart(self._iw_dof, self._iw_scale).rvs(num_samples)
        idx1, idx2 = np.triu_indices(self._ncols)
        return cov[:, idx1, idx2]

    def _assemble_covs(self, samples):
        covs = np.zeros((samples.shape[0], self._ncols, self._ncols))
        idx1, idx2 = np.triu_indices(self._ncols)
        covs[:, idx1, idx2] = samples
        covs += np.transpose(np.triu(covs, 1), axes=(0, 2, 1))
        return covs

    @staticmethod
    def _is_pos_semidef(x):
        try:
            np.linalg.cholesky(x)
            return 1
        except np.linalg.linalg.LinAlgError:
            return 0

    def _check_samples_shape(self, cov):
        if len(cov.shape) != 2 or cov.shape[1] != self._dim:
            raise ValueError('Covariance must be 3D array w/ index 0 '
                             'representing the sample index and indices 1 and '
                             '2 representing the 2D covariance indices (which '
                             'should also be equal size)')


'''
Notices:
Copyright 2018 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRessED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNess FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLess THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.
'''


from .particles import Particles
from ..mcmc.kernel_base import MCMCKernel
from copy import copy
import numpy as np


class Mutator:
    '''
    Mutates particles using an MCMC kernel.
    '''
    def __init__(self, mcmc_kernel):
        '''
        :param mcmc_kernel: a kernel object for conducting particle mutation
        :type mcmc_kernel: MCMCKernel object
        '''
        self.mcmc_kernel = mcmc_kernel

    def mutate(self, particles, phi, num_samples):
        cov = particles.compute_covariance()
        mutated = self.mcmc_kernel.mutate_particles(particles.param_dict,
                                                    particles.log_likes,
                                                    num_samples,
                                                    cov, phi)
        new_particles = Particles(mutated[0], mutated[1], particles.log_weights)
        new_particles._total_unlw = particles.total_unnorm_log_weight
        new_particles.attrs.update({'phi': phi})
        return new_particles

    @property
    def mcmc_kernel(self):
        return self._mcmc_kernel

    @mcmc_kernel.setter
    def mcmc_kernel(self, mcmc_kernel):
        if not isinstance(mcmc_kernel, MCMCKernel):
            raise TypeError
        self._mcmc_kernel = mcmc_kernel
